struct proto_iterator_tag {};

#define IS_A_PROTO_ITERATOR(C,N) \
    typedef proto_iterator_tag iterator_tag; \
    typedef tuple<N> key_type

#define PROTO_ITERATOR_METHODS(C,N) \
    const key_type operator *() const { return v_deref(); } \
    inline bool operator ==(const C &y) { return v_equal(y); } \
    bool operator !=(const C &y) { return !v_equal(y); } \
    C &operator=(const C &x) { v_set(x); return *this; }

// The following methods must be implemented,
//     const key_type v_deref() const;
//     bool v_equal(const C &) const;
//     void v_set(const C &);

template <class I>
inline typename I::iterator_tag bsp_iterator_category(const I&) {
    typedef typename I::iterator_tag category;
    return category();
}

///////////////////////////////////////////////////////////////////////////

template<class ITER>
int count_impl(const ITER &first, const ITER &last,
	       const proto_iterator_tag &);

template<class ITER>
int count(const ITER &first, const ITER &last)
{
    return count_impl(first, last, bsp_iterator_category(first));
}

///////////////////////////////////////////////////////////////////////////

template<class ITER, class KEY>
ITER search_impl(const ITER &first, const ITER &last,
	    const proto_iterator_tag &, KEY k);

template<class ITER, class KEY>
ITER search(const ITER &first, const ITER &last, KEY k)
{
    return search_impl(first, last, bsp_iterator_category(first), k);
}

///////////////////////////////////////////////////////////////////////////

struct unordered_iterator_tag : public virtual proto_iterator_tag {};

#define IS_AN_UNORDERED_ITERATOR(C,N) \
    typedef unordered_iterator_tag iterator_tag; \
    typedef tuple<N> key_type

#define UNORDERED_ITERATOR_METHODS(C,N) \
    void operator ++(int) { v_incr(); } \
    void operator ++() { v_incr(); } \
    PROTO_ITERATOR_METHODS(C,N)

// The following methods must be implemented,
//     void v_incr();

template<class ITER>
int count_impl(const ITER &first, const ITER &last,
		const unordered_iterator_tag &)
{
    int num = 0;
    ITER it; for (it = first; it != last;  it++) num++;
    return num;
}

template<class ITER, class KEY>
ITER search_impl(const ITER &first, const ITER &last,
		 const unordered_iterator_tag &, KEY k)
{
    ITER it; for (it = first; it != last && *it != k;  it++) ;
    if (*it == k) return it;
    else { it = last; return it; }
}

///////////////////////////////////////////////////////////////////////////

struct increasing_iterator_tag : public virtual proto_iterator_tag {};

#define IS_AN_INCREASING_ITERATOR(C,N) \
    typedef increasing_iterator_tag iterator_tag; \
    typedef tuple<N> key_type

#define INCREASING_ITERATOR_METHODS(C,N) \
    INCREASING_ITERATOR_METHODS_(C,N); \
    UNORDERED_ITERATOR_METHODS(C,N)

#define INCREASING_ITERATOR_METHODS_(C,N)
    /* UNORDERED_ITERATOR_METHODS already includes all of the methods. */

template<class ITER>
int count_impl(const ITER &first, const ITER &last,
	       const increasing_iterator_tag &)
{
    int num = 0;
    ITER it; for (it = first; it != last;  it++) num++;
    return num;
}

template<class ITER, class KEY>
ITER search_impl(const ITER &first, const ITER &last,
		 const increasing_iterator_tag &, KEY k)
{
    ITER it(first); 
    for ( ; 
	 it != last && *it < k;  
	 it++) ;
    if (*it == k) return it;
    else { it = last; return it; }
}

///////////////////////////////////////////////////////////////////////////

struct decreasing_iterator_tag : public virtual proto_iterator_tag {};

#define IS_A_DECREASING_ITERATOR(C,N) \
    typedef decreasing_iterator_tag iterator_tag; \
    typedef tuple<N> key_type

#define DECREASING_ITERATOR_METHODS(C,N) \
    DECREASING_ITERATOR_METHODS_(C,N); \
    UNORDERED_ITERATOR_METHODS(C,N)

#define DECREASING_ITERATOR_METHODS_(C,N) \
    void operator --(int) { v_decr(); } \
    void operator --() { v_decr(); }

// The following methods must be implemented,
//     void v_decr();

///////////////////////////////////////////////////////////////////////////

struct ordered_iterator_tag :
    public increasing_iterator_tag,
    public decreasing_iterator_tag {};

#define IS_AN_ORDERED_ITERATOR(C,N) \
    typedef ordered_iterator_tag iterator_tag; \
    typedef tuple<N> key_type

#define ORDERED_ITERATOR_METHODS(C,N) \
    INCREASING_ITERATOR_METHODS_(C,N); \
    DECREASING_ITERATOR_METHODS_(C,N); \
    UNORDERED_ITERATOR_METHODS(C,N)

///////////////////////////////////////////////////////////////////////////

struct offset_iterator_tag :
    public ordered_iterator_tag {};

#define IS_AN_OFFSET_ITERATOR(C,N) \
    typedef offset_iterator_tag iterator_tag; \
    typedef tuple<N> key_type

#define OFFSET_ITERATOR_METHODS(C,N) \
    void v_incr() { v_incr_delta(1); } \
    void v_decr() { v_decr_delta(1); } \
    int operator -(const C &y) { return v_distance(y); } \
    void operator +=(int d) { v_incr_delta(d); } \
    void operator -=(int d) { v_decr_delta(d); } \
    ORDERED_ITERATOR_METHODS(C,N)

// The following methods must be implemented,
//     int v_distance(const C &);
//     void v_incr_delta(int d);
//     void v_decr_delta(int d);

template<class ITER>
int count_impl(const ITER &first, const ITER &last,
	       const offset_iterator_tag &)
{
    return last - first;
}

///////////////////////////////////////////////////////////////////////////

struct interval_iterator_tag : public offset_iterator_tag {};

class interval_iterator
{
public:
    typedef interval_iterator_tag iterator_tag;
    typedef tuple<_1> key_type;
protected:
    key_type k;
public:
    interval_iterator() : k(-1) { }
    interval_iterator(key_type k) : k(k) { }
    const key_type v_deref() const { return k; }
    bool v_equal(const interval_iterator &y) const
	{ return k == y.k; }
    void v_set(const interval_iterator &y)
	{ k = y.k; }
    int v_distance(const interval_iterator &y) {
	key_type result = k;
	result -= y.k;
	return result; }
    void v_incr_delta(int d) { k += tuple<_1>(d); }
    void v_decr_delta(int d) { k -= tuple<_1>(d); }
    OFFSET_ITERATOR_METHODS(interval_iterator,_1);
};

template<class ITER, class KEY>
ITER search_impl(const ITER &first, const ITER &last,
		 const interval_iterator_tag &, KEY k)
{
    if (*first <= k && k < *last)
	return ITER(k);
    else
	return last;
}
